:- [sequents, recurrence, gensym_simple].

apply_rules(Formula, Sequent, Branches) :-
	try_to_apply(Formula, Sequent, Branches),
	remove_if_recur(Branches).


remove_if_recur(Branches) :-
	sets_recur(Branches),
	!,
	remove_sequents(Branches),
	fail.

remove_if_recur(_).


try_to_apply(Formula, Sequent, Branches) :-
	apply_or2(Formula, Sequent, Branches).

try_to_apply(Formula, Sequent, Branches) :-
	apply_or3(Formula, Sequent, Branches).

try_to_apply(Formula, Sequent, Branches) :-
	apply_always(Formula, Sequent, Branches).

try_to_apply(Formula, Sequent, Branches) :-
	apply_eventually(Formula, Sequent, Branches).
	
try_to_apply(Formula, Sequent, Branches) :-
	apply_db(Formula, Sequent, Branches).

try_to_apply(Formula, Sequent, Branches) :-
	apply_always_eventually(Formula, Sequent, Branches).


apply_or2(always Conjunction,
	   Sequent, [Branch1, Branch2]) :-
	Conjunction =.. [or, Literal1, Literal2],
	literal(Literal1),
	literal(Literal2),

	formula_not_included(Sequent, Literal1),
	formula_not_included(Sequent, Literal2),

	generate_branch(Sequent, Branch1),
	copy(Sequent, Branch1),
	remove(Branch1, always Conjunction),%
	supplement_sequent(Branch1, Literal1),
	mark_applied_or(Branch1),
	
	generate_branch(Sequent, Branch2),
	copy(Sequent, Branch2),
	remove(Branch2, always Conjunction),%
	supplement_sequent(Branch2, Literal2),
	mark_applied_or(Branch2).

apply_or3(always Conjunction,
	   Sequent, [Branch1, Branch2, Branch3]) :-
	Conjunction =.. [or, Literal1, Conjunction2],
	Conjunction2 =.. [or, Literal2, Literal3],
	literal(Literal1),
	literal(Literal2),
	literal(Literal3),

	formula_not_included(Sequent, Literal1),
	formula_not_included(Sequent, Literal2),
	formula_not_included(Sequent, Literal3),
	
	generate_branch(Sequent, Branch1),
	copy(Sequent, Branch1),
	remove(Branch1, always Conjunction),%
	supplement_sequent(Branch1, Literal1),
	mark_applied_or(Branch1),
	
	generate_branch(Sequent, Branch2),
	copy(Sequent, Branch2),
	remove(Branch2, always Conjunction),%
	supplement_sequent(Branch2, Literal2),
	mark_applied_or(Branch2),

	generate_branch(Sequent, Branch3),
	copy(Sequent, Branch3),
	remove(Branch3, always Conjunction),%
	supplement_sequent(Branch3, Literal3),
	mark_applied_or(Branch3).


apply_always(always Conjunction,
	       Sequent, [Branch1, Branch2]) :-
	Conjunction =.. [or, Literal1, always Literal2],
	literal(Literal1),
	literal(Literal2),

	or_not_applied(Sequent),

	formula_not_included(Sequent, Literal1),
	formula_not_included(Sequent, always Literal2),
	
	generate_branch(Sequent, Branch1),
	copy(Sequent, Branch1),
	supplement_sequent(Branch1, Literal1),
 	mark_applied_always_left(Literal1),
	
	generate_branch(Sequent, Branch2),
	copy(Sequent, Branch2),
	remove(Branch2, always Conjunction),
	supplement_sequent(Branch2, always Literal2).
	

apply_eventually(always Conjunction,
		   Sequent, [Branch1, Branch2]) :-
	Conjunction =.. [or, Literal1, eventually Literal2],
	literal(Literal1),
	literal(Literal2),

	or_not_applied(Sequent),
	always_left_not_applied(Sequent),

	formula_not_included(Sequent, always Literal1),
	formula_not_included(Sequent, Literal2),
	
	generate_branch(Sequent, Branch1),
	copy(Sequent, Branch1),
	remove(Branch1, always Conjunction),
	supplement_sequent(Branch1, always Literal1),
	
	generate_branch(Sequent, Branch2),
	copy_with_always(Sequent, Branch2),
	supplement_sequent(Branch2, Literal2).


apply_db(always Conjunction,
	   Sequent, [Branch1, Branch2, Branch3]) :-
	Conjunction =.. [or, Literal1, eventually Literal2],
	literal(Literal1),
	literal(Literal2),

	or_not_applied(Sequent),
	always_left_not_applied(Sequent),

	formula_not_included(Sequent, Literal1),
	formula_not_included(Sequent, always eventually Literal2),
	
	generate_branch(Sequent, Branch1),
	copy(Sequent, Branch1),
	supplement_sequent(Branch1, Literal1),
	
	generate_branch(Sequent, Branch2),
	copy_with_always(Sequent, Branch2),
	remove(Branch2, always Conjunction),
	supplement_sequent(Branch2, always eventually Literal2),

	generate_branch(Sequent, Branch3),
	copy_with_always(Sequent, Branch3),
	remove(Branch3, always Conjunction),
	supplement_sequent(Branch3, Literal2),
	supplement_sequent(Branch3, always(Literal2 or always nott Literal2)),
	( Literal1 @=< Literal2 ->
	    supplement_sequent(Branch3, always(Literal1 or Literal2));
	    supplement_sequent(Branch3, always(Literal2 or Literal1))
	).


apply_always_eventually(always eventually Literal,
			  Sequent, [Branch]) :-
	literal(Literal),

	or_not_applied(Sequent),
	always_left_not_applied(Sequent),
	
	formula_not_included(Sequent, Literal),
	
	generate_branch(Sequent, Branch),
	copy_with_always(Sequent, Branch),
	supplement_sequent(Branch, Literal).


literal(Variable) :-
	atomic(Variable).

literal(nott Variable) :-
	atomic(Variable).


generate_branch(Root, Branch) :-
	gensym_simple(a, Node),
	Branch =.. [Node, Root].


mark_applied_or(Sequent) :-
	recorda(Sequent, applied_or).


or_not_applied(Sequent) :-
	recorded(Sequent, applied_or),
	!,
	fail.

or_not_applied(_).


mark_applied_always_left(Sequent) :-
	recorda(Sequent, applied_always_left).


always_left_not_applied(Sequent) :-
	recorded(Sequent, applied_always_left),
	!,
	fail.

always_left_not_applied(_).
	

formula_not_included(Sequent, Formula) :-
	recorded(Sequent, Formula),
	!,
	fail.

formula_not_included(_, _).

